"""Variables"""
import yaml

class Variables:
    """Variables"""

    def __init__(self,
                 name=None,
                 p0=0.7, pA=0, pS=0.1, ap=0.42, bn=0.53,
                 Kc=10, Rmax=100, pdT=0.5, pdI=0.5,
                 yPC=0.55, yQ=0.4, yI=0.7, kPC=0.8, mu=0.4, ics=0.4,
                 max_immune_cell_count=2000,
                 kQ=0.4, kI=0.6, ci=0.5, PK=1, max_energy_level=30,
                 necrotic_distance=30, quiescent_distance=30,
                 treatment_start_time=10, injection_interval=10,
                 time_constant=3, drug_concentration=0.1
                 ) -> None:
        self.name = name

        # Static variables
        self.p0 = p0
        self.pA = pA
        self.pS = pS
        self.mu = mu

        self.ics = ics
        self.max_immune_cell_count = max_immune_cell_count

        self.ap = ap
        self.bn = bn

        self.Kc = Kc
        self.Rmax = Rmax

        self.pdT = pdT
        self.pdI = pdI

        self.yPC = yPC
        self.yQ = yQ
        self.yI = yI
        self.kPC = kPC
        self.kQ = kQ
        self.kI = kI
        self.ci = ci
        self.PK = PK

        self.max_energy_level = max_energy_level
        self.necrotic_distance = necrotic_distance
        self.quiescent_distance = quiescent_distance

        self.treatment_start_time = treatment_start_time
        self.injection_interval = injection_interval
        self.time_constant = time_constant
        self.drug_concentration = drug_concentration

        # Dynamic variables
        self.Rt = 0
        self.injection_number = 0
        self.time_delta = 5
        self.time = 0

    @property
    def Wp(self) -> float:
        return self.ap * self.Rt ** (2/3)

    @property
    def Rn(self) -> float:
        return self.Rt - self.bn * self.Rt ** (2/3)

    @property
    def days_elapsed(self) -> int:
        return self.time // 24

    def time_step(self):
        self.time += self.time_delta

    @property
    def is_treatment(self) -> bool:
        """Return true if chemotherapy injection is in process"""

        # Treatment has not started yet
        if self.days_elapsed < self.treatment_start_time:
            return False

        days_from_start = self.days_elapsed - self.treatment_start_time

        # No drug in blood
        if days_from_start % self.injection_interval > self.time_constant:
            return False

        return True

    @property
    def is_injection_start(self) -> bool:
        """Return true if it is a day of injection"""

        if not self.is_treatment:
            return False

        days_from_start = self.days_elapsed - self.treatment_start_time

        if days_from_start % self.injection_interval == 0 and self.time % 24 == 0:
            return True
        return False


class ConfigFileException(Exception):
    """Exception for incorrect config file"""


def read_variables(filepath: str) -> list[Variables]:
    """
    Read configuration file and return list of Variables for
    each proposed simulation 
    """

    config = {}
    with open(filepath, 'r', encoding="utf8") as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError:
            raise ValueError("You passed incorrect yaml file")

        variables = []

        global_variables = config.get('global')
        if not global_variables:
            raise ConfigFileException("You must define `global` section in you config file")

        simulation_list = config.get("simulations")
        if not simulation_list:
            raise ConfigFileException("You must define `simulations` section in you config file")

        for simulation_identifier, simulation_config in simulation_list.items():
            simulation_name = simulation_config.get('name')

            if not simulation_name:
                raise ConfigFileException(f"You must define `name` in simulation\
                                           {simulation_identifier}")

            simulation_variables = global_variables | simulation_config

            if len(variables) != 4:
                variables.append(Variables(**simulation_variables))
            else:
                raise ConfigFileException("You can set 4 simulations at most")

    return variables



"""Cell"""


class Cell:
    """
    Cell - entity that can contain either biological cell
    or other entity
    """

    __dict__ = [
        "x",
        "y",
        "_entity",
        "neighbors",
        "add_entity_callback",
        "remove_entity_callback",
    ]

    def __init__(self, x: int, y: int, entity=None) -> None:
        """Initialize cell"""
        self.x = x
        self.y = y
        self._entity = entity
        self.neighbors = []
        self.distance = 0
        self.phi = 0

        self.add_entity_callback = None
        self.remove_entity_callback = None

    def __hash__(self):
        return hash((self.x, self.y))

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    @property
    def entity(self):
        """Return entity"""
        return self._entity

    @entity.setter
    def entity(self, entity_) -> None:
        if self.empty and entity_ is not None:
            self.add_entity_callback(self)

        if entity_ is None and not self.empty:
            self.remove_entity_callback(self)

        self._entity = entity_

    @property
    def empty(self) -> bool:
        """Return true if cell is empty"""
        return self.entity is None

    @property
    def entity_id(self) -> int:
        """Return identifier of the entity"""
        return self.entity.ID if self.entity else 0

    @property
    def color(self):
        if self.entity:
            return self.entity.color
        return 0, 0, 0

    def get_free_neighbor(self):
        """Return empty cell"""
        return [cell for cell in self.neighbors if cell.empty]



"""Grid and cell"""

from typing import Iterable
import numpy as np
from random import randint
from math import sqrt
from .cell import Cell
from .entity import Entity


class Grid:
    """Grid"""

    def __init__(self, width=1000, height=1000) -> None:
        """Initialize the grid"""

        self.width = width
        self.height = height

        self.active_cells = set()
        self.grid = [[Cell(x, y) for x in range(width)] for y in range(height)]

        for row in self.grid:
            for cell in row:
                cx, cy = self.center
                cell.add_entity_callback = lambda c: self.add_active_cell(c)
                cell.remove_entity_callback = lambda c: self.remove_active_cell(c)
                cell.neighbors = self.get_neighbors_of(cell)
                cell.distance = sqrt((cell.x - cx) ** 2 + (cell.y - cy) ** 2)
                cell.phi = np.arctan2(cell.y - cy, cell.x - cx)

    @property
    def center(self) -> tuple[int, int]:
        """Return coordinate of center"""
        return (self.width // 2, self.height // 2)

    @property
    def cells(self) -> Iterable:
        """Return all cells iterator"""
        return self.active_cells

    @property
    def coloured_cells(self) -> Iterable:
        return [(cell.x, cell.y, cell.entity.color) for cell in self.active_cells]

    def add_active_cell(self, cell: Cell) -> None:
        """Add active cell"""
        self.active_cells.add(cell)

    def remove_active_cell(self, cell: Cell) -> None:
        """Add active cell"""
        self.active_cells.remove(cell)

    def place_entity(self, entity: Entity, x: int, y: int) -> None:
        """Place entity on grid by coordinates"""
        if not -self.width < x < self.width or not -self.height < y < self.height:
            raise ValueError("You cannot put an entity outside the boundaries")

        self.grid[y][x].entity = entity
        self.add_active_cell(self.grid[y][x])

    def to_array(self) -> list[list[int]]:
        """
        Convert list of Cell objects to list of int
        """
        return [[cell.entity_id for cell in row] for row in self.grid]

    def get_neighbors_of(self, cell: Cell) -> list[Cell]:
        """
        Return neighbors of given cell - all adjacent cells
        in the grid
        """

        neighbors = []
        for dx in range(-1, 2):
            for dy in range(-1, 2):
                if dx == dy == 0:
                    continue

                x = (cell.x + dx) % self.width
                y = (cell.y + dy) % self.height

                neighbors.append(self.grid[y][x])

        return neighbors

    def get_random_free_cell(self) -> Cell:
        """Return random free cell"""
        x, y = randint(0, self.width - 1), randint(0, self.height - 1)
        while not self.grid[x][y].empty:
            x, y = randint(0, self.width - 1), randint(0, self.height - 1)
        return self.grid[x][y]



"""global constants"""

SCREEN_HEIGHT = 900
SCREEN_WIDTH = 1600

BUTTON_HEIGHT = 100
BUTTON_WIDTH = 100
BUTTON_ADD = "./assets/icons8-add-100.png"
BUTTON_START = "./assets/icons8-watermelon-100.png"
BUTTON_STOP = "./assets/icons8-cancel-138.png"

BLOCK_SIZE_DIVISIBLE = 500



"""Entities"""
import math
import numpy as np
from random import random, choice
from .variables import Variables
from .cell import Cell


MAX_PROLIFERATION_POTENTIAL = 30
MAX_PROLIFERATION_COLOR = np.array([250,0,0])
LOW_PROLIFERATION_COLOR = np.array([0, 0, 0])

DELTA = ( MAX_PROLIFERATION_COLOR - LOW_PROLIFERATION_COLOR ) / MAX_PROLIFERATION_POTENTIAL


def get_chemo_death_probability(theta, k, y, variables: Variables) -> float:
    """Return probability of cell's death due to chemotherapy"""
    l = k * variables.drug_concentration / (theta * y * variables.injection_number + 1)
    return l * variables.PK * math.e ** ( -variables.ci * (variables.days_elapsed - \
            variables.injection_number * variables.time_constant))


class Entity:
    """Entity"""

    __dict__ = ["cell", "neighbors", "free_neighbors", "variables"]

    def __init__(self) -> None:
        """Initialize entity"""
        self.cell = None
        self.neighbors = None
        self.free_neighbors = None
        self.variables = None

    def next_state(self) -> None:
        """Mutate cell or neighbors to represent the next state"""

    def move_to(self, next_cell: Cell) -> None:
        """Move from current cell to next"""
        self.cell.entity = None
        next_cell.entity = self

    def move_to_random(self) -> None:
        """Move to random free neighbor"""
        free_neighbor = self.free_neighbors

        if not self.cell.entity: # cell has died
            return

        if not free_neighbor:
            return

        cell = choice(free_neighbor)
        self.move_to(cell)


class BiologicalCell(Entity):
    """Biological cell"""

    ID = 1

    __dict__ = ["ID", "proliferation_potential", "cell", "neighbors", "free_neighbors", "variables"]

    def __init__(self, proliferation_potential=MAX_PROLIFERATION_POTENTIAL,
                 *args, **kwargs) -> None:
        """Initialize Biological cell"""
        super().__init__(*args, **kwargs)

        self.proliferation_potential = proliferation_potential
        self.energy_level = 0

    @property
    def apotisis_probability(self) -> float:
        """Return probability of spontaneous death"""
        return self.variables.pA

    @property
    def proliferation_probability(self) -> float:
        """Return probability of proliferation"""
        return self.variables.p0 * (1 - (self.cell.distance / (self.variables.Rmax - self.variables.Kc) ))

    @property
    def migration_probability(self) -> float:
        """Return probability of migration"""
        return self.variables.mu

    def next_state(self, *random_values) -> None:
        """Next state implementation to BiologicalCell"""
        pass

    def proliferate(self) -> None:
        """Proliferate"""
        free_neighbors = self.free_neighbors
        if not free_neighbors:
            return

        free_cell = choice(free_neighbors)

        if self.proliferation_potential <= 0:
            self.apotose()
            return

        free_cell.entity = self.replicate()

    def replicate(self) -> Entity:
        """Return daughter cell"""
        daughter =  BiologicalCell(self.proliferation_potential - 1)
        self.proliferation_potential -= 1
        return daughter

    def apotose(self) -> None:
        """Cell death"""
        self.cell.entity = None

    @property
    def color(self):
        r, g, b =  LOW_PROLIFERATION_COLOR + self.proliferation_potential * DELTA
        return int(r), int(g), int(b)


class CancerCell(BiologicalCell):
    """Cancer cell"""

    def __init__(self, proliferation_potential=MAX_PROLIFERATION_POTENTIAL,
                 *args, **kwargs) -> None:
        super().__init__(proliferation_potential, *args, **kwargs)

    def next_state(self, variables: Variables, *random_variables) -> None:
        apotosis, proliferation, migration, theta, death = random_variables

        if apotosis <= self.apotisis_probability:
            self.apotose()
            return

        if proliferation <= self.proliferation_probability:
            self.proliferate()

        if migration <= self.migration_probability:
            self.move_to_random()

        self.process_chemotherapy(variables=variables, theta=theta, death=death)

        if self.energy_level >= variables.quiescent_distance:
            self.cell.entity = QuiescentCell(self)

        if self.energy_level >= variables.necrotic_distance:
            self.cell.entity = NecroticCell()


    def process_chemotherapy(self, variables: Variables, theta: float, death: float) -> None:
        """Process the effect of the chemotherapy"""

        if not variables.is_treatment:
            return

        chemo_death_probability = self.get_chemo_death_probability(theta=theta, variables=variables)

        if death <= chemo_death_probability:
            self.apotose()

    def get_chemo_death_probability(self, variables, theta) -> float:
        """Return the probability of death due to chemotherapy"""
        return   get_chemo_death_probability(theta=theta,
                                             k=variables.kPC,
                                             y=variables.yPC,
                                             variables=variables)

    def replicate(self) -> Entity:
        """Return daughter cell"""
        daughter =  CancerCell(self.proliferation_potential - 1)
        self.proliferation_potential -= 1
        return daughter

    @property
    def color(self) -> tuple[int, int, int]:
        r, g, b =  LOW_PROLIFERATION_COLOR + self.proliferation_potential * DELTA
        return int(r), int(g), int(b)


class QuiescentCell(BiologicalCell):
    """Quiescent cell"""

    def __init__(self, previous_entity, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.previous_entity = previous_entity

    def next_state(self, variables: Variables, *random_variables) -> None:
        if self.energy_level < self.variables.quiescent_distance:
            self.cell.entity = self.previous_entity

    @property
    def color(self):
        return self.previous_entity


class NecroticCell(BiologicalCell):
    """Necrotic cell"""

    def next_state(self, variables: Variables, *random_variables) -> None:
        pass

    @property
    def color(self):
        return (133, 21, 21)


class TrueStemCell(CancerCell):
    """
    True Stem cell.
    Cell that is immortal and can give birth to either
    RTC or other True stem cell
    """

    ID = 3

    @property
    def apotisis_probability(self) -> float:
        """Return probability of spontaneous death"""
        return 0

    def replicate(self) -> Entity:
        """Return daughter cell"""
        new_stem_chance = random()
        if new_stem_chance <= self.variables.pS:
            daughter = TrueStemCell(self.proliferation_potential)
        else:
            daughter =  CancerCell(self.proliferation_potential)
        return daughter

    def get_chemo_death_probability(self, variables, theta) -> float:
        return 0

    @property
    def color(self):
        return 255, 238, 0


class ImmuneCell(BiologicalCell):
    """Immune cell"""

    def next_state(self, variables: Variables, *random_values) -> None:
        *_, cancer_cell_death_probability, immune_cell_death_probability = random_values

        self.move_to_random(immune_cell_death_probability)

        for neighbor in self.neighbors:
            if neighbor.empty:
                continue

            if isinstance(neighbor.entity, CancerCell):
                if cancer_cell_death_probability <= variables.pdT:
                    neighbor.entity = None

                if immune_cell_death_probability <= variables.pdI:
                    self.apotose()

                break

    def move_to_random(self, direction_probability) -> None:
        """Move to random free neighbor"""

        if direction_probability <= self.variables.ics:
            free_neighbor = sorted(self.free_neighbors, key=lambda c: c.distance)[:3]
        else:
            free_neighbor = sorted(self.free_neighbors, key=lambda c: c.distance)

        if not self.cell.entity: # cell has died
            return

        if not free_neighbor:
            return

        cell = choice(free_neighbor)
        self.move_to(cell)

    @property
    def color(self):
        return 245, 91, 209
